# coding=utf-8

import logging
from typing import TYPE_CHECKING, Dict, Any
from ehforwarderbot import EFBChat, ChatType
from fbchat.models import Thread, User, Page, Group, Room

from .utils import get_value

if TYPE_CHECKING:
    from . import FBMessengerChannel
    from .efms_client import EFMSClient


class EFMSChat(EFBChat):

    logger = logging.getLogger("EFMSChat")

    cache = dict()

    def __init__(self, channel: 'FBMessengerChannel', thread: Thread=None,
                 graph_ql_thread: Dict[str, Any]=None,
                 uid: str=None, lazy_load=False):
        """
        Create a chat based on fbchat Thread objects or GraphQL result dict.

        Args:
            channel: The EFMS channel object
            thread: fbchat Thread object as information source
            graph_ql_thread: GraphQL thread as information source
            uid: Thread ID, and retrieve necessary details from server
            lazy_load: If chat info need to be loaded later
        """
        super().__init__(channel)
        self.channel = channel
        self.thread = thread
        self.chat_uid = uid
        self.client: EFMSClient = channel.client
        self.loaded = False
        self.graph_ql_thread = graph_ql_thread

        # Mark shelf
        if self.chat_uid == self.client.uid:
            self.self()
            return
        if not lazy_load:
            self.load_chat()

    def load_chat(self):
        self.loaded = True

        thread: Thread = self.thread

        if self.is_self:
            return

        if self.graph_ql_thread:
            return self.load_graph_ql_thread()

        if not thread:
            if self.chat_uid in EFMSChat.cache:
                self.clone(EFMSChat.cache[self.chat_uid])
                return
            thread = self.client.fetchThreadInfo(self.chat_uid)[str(self.chat_uid)]
            self.thread = thread

        self.logger.debug("Parsing fbchat thread object: %s", thread)

        self.chat_uid = thread.uid
        self.chat_name = thread.name
        self.vendor_specific['chat_type'] = thread.type.name.capitalize()

        if self.chat_uid == self.client.uid:
            self.self()
            EFMSChat.cache[self.chat_uid] = self
            return

        if isinstance(thread, User):
            self.chat_type = ChatType.User
            self.chat_alias = thread.own_nickname or thread.nickname
        elif isinstance(thread, Page):
            self.chat_type = ChatType.User
        elif isinstance(thread, Group) or isinstance(thread, Room):
            self.chat_type = ChatType.Group

            # Lazy load if the group has a name,
            # otherwise members' names are needed to build group name

            lazy_load = bool(self.chat_name)
            names = []

            if self.is_chat:
                for i in thread.participants:
                    member = EFMSChat(self.channel, uid=i, lazy_load=lazy_load)
                    if thread.nicknames and i in thread.nicknames:
                        member.chat_alias = thread.nicknames[str(i)]
                    if not lazy_load:
                        names.append(member.chat_alias or member.chat_name)
                    member.is_chat = False
                    self.members.append(member)

            if not lazy_load:
                if names:
                    names.sort()
                    self.chat_name = ", ".join(names[:3])
                    if len(names) > 3:
                        self.chat_name += ", and %d more" % (len(names) - 3)
                else:
                    self.chat_name = "Group %s" % self.chat_uid
        else:
            self.chat_type = ChatType.System
        EFMSChat.cache[self.chat_uid] = self

    def load_graph_ql_thread(self, recursive=True):
        self.logger.debug('Parsing chat as GraphQL dict: %s', self.graph_ql_thread)

        if 'thread_key' in self.graph_ql_thread:
            # Data of a thread
            self.logger.debug('GraphQL information provided is a thread.')
            self.chat_name = self.graph_ql_thread['name']
            if self.graph_ql_thread['thread_type'] == 'ONE_TO_ONE':
                self.chat_type = ChatType.User
                self.chat_uid = self.graph_ql_thread['thread_key']['other_user_id']
                item = None
                for i in self.graph_ql_thread['all_participants']['nodes']:
                    if i['messaging_actor']['id'] == self.chat_uid:
                        item = i['messaging_actor']
                        break
                self.logger.debug("[%s] One to one member info: %s", self.chat_uid, item)
                if item:
                    self.chat_name = self.chat_name or item['name']
                    self.vendor_specific['profile_picture_url'] = item['big_image_src']['uri']
                    self.vendor_specific['chat_type'] = item.get('__typename')
            elif self.graph_ql_thread['thread_type'] == 'GROUP':
                self.chat_type = ChatType.Group
                self.chat_uid = self.graph_ql_thread['thread_key']['thread_fbid']
                self.vendor_specific['chat_type'] = 'Group'
                self.vendor_specific['profile_picture_url'] = \
                    get_value(self.graph_ql_thread, ('image', 'uri'))
                for i in self.graph_ql_thread['all_participants']['nodes']:
                    self.members.append(EFMSChat(self.channel,
                                                 graph_ql_thread=i,
                                                 lazy_load='name' in i['messaging_actor']))
                    self.members[-1].group = self
                    self.members[-1].is_chat = False
                if not self.chat_name:
                    names = sorted(i.chat_name for i in self.members)
                    self.chat_name = ", ".join(names[:3])
                    if len(names) > 3:
                        self.chat_name += ", and %s more" % (len(names) - 3)
        elif 'messaging_actor' in self.graph_ql_thread:
            self.logger.debug('GraphQL information provided is a thread member.')
            # data of a group member
            self.chat_uid = self.graph_ql_thread['messaging_actor']['id']
            if 'name' in self.graph_ql_thread['messaging_actor']:
                self.logger.debug('[%s] Thread member information is complete.', self.chat_uid)
                self.chat_name = self.graph_ql_thread['messaging_actor']['name']
                self.chat_type = ChatType.User
                self.vendor_specific['chat_type'] = self.graph_ql_thread['messaging_actor'].get('__typename')
                self.vendor_specific['profile_picture_url'] = \
                    get_value(self.graph_ql_thread, ('messaging_actor', 'big_image_src', 'uri'))
            elif recursive:
                self.logger.debug('[%s] Thread member information is incomplete.', self.chat_uid)
                self.graph_ql_thread = self.client.get_thread_info(self.chcat_uid)
                return self.load_graph_ql_thread(recursive=False)

        if self.chat_uid == self.client.uid:
            self.self()

        EFMSChat.cache[self.chat_uid] = self

    def clone(self, other: 'EFMSChat'):
        self.__dict__.update(**other.__dict__)

    def __getattribute__(self, item):
        if item in {"chat_name", "chat_alias", "members", "chat_type"} and not self.loaded:
            self.load_chat()
        return super().__getattribute__(item)



# coding=utf-8

import logging
import time
import threading
import emoji
from typing import TYPE_CHECKING, Set

from fbchat.models import Thread, Message, TypingStatus, ThreadType, Mention, EmojiSize, MessageReaction
from ehforwarderbot import EFBMsg, MsgType
from ehforwarderbot.message import EFBMsgLinkAttribute, EFBMsgStatusAttribute, ChatType
from ehforwarderbot.exceptions import EFBMessageTypeNotSupported

from .utils import get_value

if TYPE_CHECKING:
    from . import FBMessengerChannel


class MasterMessageManager:

    logger = logging.getLogger("MasterMessageManager")

    def __init__(self, channel: 'FBMessengerChannel'):
        self.channel = channel
        channel.supported_message_types: Set[MsgType] = {MsgType.Text, MsgType.Image, MsgType.Sticker,
                                                         MsgType.Audio, MsgType.File, MsgType.Video,
                                                         MsgType.Status, MsgType.Unsupported}
        self.client = channel.client
        self.flag = channel.flag

    def send_message(self, msg: EFBMsg) -> EFBMsg:
        self.logger.debug("Received message from master: %s", msg)

        try:
            target_msg_offset = 0
            prefix = ""

            mentions = []

            # Send message reaction
            if msg.target and msg.text.startswith('r`') and getattr(MessageReaction, msg.text[2:], None)\
                    and msg.uid.startswith("mid.$"):
                self.logger.debug("[%s] Message is a reaction to another message: %s", msg.uid, msg.text)
                msg_id = ".".join(msg.target.uid.split(".", 2)[:2])
                self.client.reactToMessage(msg_id, getattr(MessageReaction, msg.text[2:]))
                msg.uid = "__reaction__"
                return msg

            # Target message
            if msg.target:
                self.logger.debug("[%s] Message replying to another message: %s", msg.uid, msg.target)
                if msg.target.chat.chat_type == ChatType.Group:
                    target_msg_name = msg.target.author.chat_alias or msg.target.author.chat_name
                    prefix = '@%s "%s"\n' % (target_msg_name, msg.target.text or msg.target.type)
                    mentions.append(Mention(msg.target.author.chat_uid, 1, len(target_msg_name)))
                else:
                    prefix = '"%s"\n' % msg.target.text or msg.target.type
                target_msg_offset = len(prefix)
                self.logger.debug("[%s] Converted to prefix: %s", msg.uid, prefix)

            # Message substitutions
            if msg.substitutions:
                self.logger.debug("[%s] Message has substitutions: %s", msg.uid, msg.substitutions)
                for i in msg.substitutions:
                    mentions.append(Mention(msg.substitutions[i].chat_uid,
                                            target_msg_offset + i[0], i[1] - i[0]))
                self.logger.debug("[%s] Translated to mentions: %s", msg.uid, mentions)

            fb_msg = Message(text=prefix + msg.text, mentions=mentions)
            thread: Thread = self.client.fetchThreadInfo(msg.chat.chat_uid)[str(msg.chat.chat_uid)]

            if msg.type in (MsgType.Text, MsgType.Unsupported):
                if msg.text[:-1] in emoji.UNICODE_EMOJI and msg.text[-1] in 'SML':
                    self.logger.debug("[%s] Message is an Emoji message: %s", msg.uid, msg.text)
                    if msg.text[-1] == 'S':
                        fb_msg.emoji_size = EmojiSize.SMALL
                    elif msg.text[-1] == 'M':
                        fb_msg.emoji_size = EmojiSize.MEDIUM
                    elif msg.text[-1] == 'L':
                        fb_msg.emoji_size = EmojiSize.LARGE
                    fb_msg.text = msg.text[:-1]
                msg.uid = self.client.send(fb_msg, thread_id=thread.uid, thread_type=thread.type)
            elif msg.type in (MsgType.Image, MsgType.Sticker):
                msg.uid = self.client.send_image_file(msg.path, msg.mime, message=fb_msg,
                                                      thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.Audio:
                file_id = self.upload_file(msg)
                msg.uid = self.client.send_audio(file_id=file_id, message=fb_msg,
                                                 thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.File:
                file_id = self.upload_file(msg)
                msg.uid = self.client.send_file(file_id=file_id, message=fb_msg,
                                                thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.Video:
                file_id = self.upload_file(msg)
                msg.uid = self.client.send_video(file_id=file_id, message=fb_msg,
                                                 thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.Status:
                attribute: EFBMsgStatusAttribute = msg.attributes
                if attribute.status_type in (EFBMsgStatusAttribute.Types.TYPING,
                                             EFBMsgStatusAttribute.Types.UPLOADING_AUDIO,
                                             EFBMsgStatusAttribute.Types.UPLOADING_VIDEO,
                                             EFBMsgStatusAttribute.Types.UPLOADING_IMAGE,
                                             EFBMsgStatusAttribute.Types.UPLOADING_FILE):
                    self.client.setTypingStatus(TypingStatus.TYPING, thread_id=thread.uid, thread_type=thread.type)
                    threading.Thread(target=self.stop_typing, args=(attribute.timeout, thread.uid, thread.type)).run()
            elif msg.type == MsgType.Link:
                attribute: EFBMsgLinkAttribute = msg.attributes
                if self.flag('send_link_with_description'):
                    info = (attribute.title,)
                    if attribute.description:
                        info += (attribute.description,)
                    info += (attribute.url,)
                    text = "\n".join(info)
                else:
                    text = attribute.url
                if fb_msg.text:
                    text = fb_msg.text + "\n" + text
                fb_msg.text = text
                msg.uid = self.client.send(fb_msg, thread_id=thread.uid, thread_type=thread.type)
            else:
                raise EFBMessageTypeNotSupported()
            return msg
        finally:
            if msg.file and not msg.file.closed:
                msg.file.close()
            self.client.markAsSeen()
            self.client.markAsRead(msg.chat.chat_uid)

    def stop_typing(self, timeout: int, thread_uid: str, thread_type: ThreadType):
        """Wait for a number of milliseconds, and stop typing."""
        time.sleep(timeout / 1000)
        self.client.setTypingStatus(TypingStatus.STOPPED, thread_id=thread_uid, thread_type=thread_type)

    def upload_file(self, msg: EFBMsg) -> int:
        """Upload media of a message as file, and return the file id."""
        response = self.client._postFile(self.client.req_url.UPLOAD, {
            'file': (msg.filename, msg.file, msg.mime)
        }, fix_request=True, as_json=True)
        return get_value(response, ('payload', 'metadata', 0, 'file_id')) or \
               get_value(response, ('payload', 'metadata', 0, 'audio_id')) or \
               get_value(response, ('payload', 'metadata', 0, 'video_id'))



from typing import TYPE_CHECKING

from fbchat import ThreadType

from .efms_chat import EFMSChat

if TYPE_CHECKING:
    from . import FBMessengerChannel


def catch_exceptions(func):
    """
    Decorator to wrap additional features.
    Return the exception string when caught.
    """
    def catch_exceptions_wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            return "Error occurred in %s(%s, %s): %r" % (func.__name__, args, kwargs, e)
    return catch_exceptions_wrap


class ExtraFunctionsManager:
    def __init__(self, channel: 'FBMessengerChannel'):
        self.channel = channel
        self.client = channel.client
        self._ = self.channel._
        self.ngettext = self.channel.ngettext

    @catch_exceptions
    def threads_list(self, args: str) -> str:
        chats = self.channel.get_chats()
        msg = self.ngettext("You have {0} thread in your thread list.",
                            "You have {0} threads in your thread list.",
                            len(chats)).format(len(chats)) + "\n"
        for i in chats:
            msg += "\n{chat.chat_uid}: {chat.chat_name} [{type}]" \
                .format(chat=i,
                        type=i.vendor_specific.get('chat_type'))
        return msg

    @catch_exceptions
    def search_users(self, args: str) -> str:
        users = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForUsers(args, limit=10)))
        msg = self.ngettext("Found {} user.",
                            "Found {} users.",
                            len(users)).format(len(users)) + "\n"
        for i in users:
            msg += "\n{chat.chat_uid}: {chat.chat_name}".format(chat=i)
        return msg

    @catch_exceptions
    def search_groups(self, args: str) -> str:
        groups = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForGroups(args, limit=10)))
        msg = self.ngettext("Found {} group.",
                            "Found {} groups.",
                            len(groups)).format(len(groups)) + "\n"
        for i in groups:
            msg += "\n{chat.chat_uid}: {chat.chat_name}".format(chat=i)
        return msg

    @catch_exceptions
    def search_pages(self, args: str) -> str:
        pages = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForPages(args, limit=10)))
        msg = self.ngettext("Found {} page.",
                            "Found {} pages.",
                            len(pages)).format(len(pages)) + "\n"
        for i in pages:
            msg += "\n{chat.chat_uid}: {chat.chat_name}".format(chat=i)
        return msg

    @catch_exceptions
    def search_threads(self, args: str) -> str:
        threads = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForThreads(args, limit=10)))
        msg = self.ngettext("Found {} thread.",
                            "Found {} threads.",
                            len(threads)).format(len(threads)) + "\n"
        for i in threads:
            msg += "\n{chat.chat_uid}: {chat.chat_name} [{type}]" \
                .format(chat=i,
                        type=i.vendor_specific.get('chat_type'))
        return msg

    @catch_exceptions
    def add_to_group(self, args: str) -> str:
        """GroupID UserID [UserID ...]"""
        args = args.split(' ')
        if len(args) < 2:
            return self._("Group ID and user IDs are required")
        self.client.addUsersToGroup(args[1:], args[0])
        return self.ngettext("User {0} are successfully added to group {1}.",
                             "Users {0} are successfully added to group {1}.",
                             len(args) - 1).format(args[1:], args[0])

    @catch_exceptions
    def remove_from_group(self, args: str) -> str:
        """GroupID UserID"""
        args = args.split(' ')
        if len(args) != 2:
            return self._("Group ID and user ID are required.")
        self.client.removeUserFromGroup(args[1], args[0])
        return self._("User {0} is successfully removed from group {1}.").format(args[1], args[0])

    @catch_exceptions
    def set_nickname(self, args: str) -> str:
        """UserID nickname"""
        args = args.split(' ', 1)
        if len(args) < 2:
            return self._("User ID and nickname are required.")
        self.client.changeNickname(args[1], args[0], args[0])
        return self._("Nickname of {0} is set to {1}.").format(*args)

    @catch_exceptions
    def set_group_title(self, args: str) -> str:
        """GroupID title"""
        args = args.split(' ', 1)
        if len(args) < 2:
            return self._("User ID and title are required.")
        self.client.changeThreadTitle(args[1], args[0], thread_type=ThreadType.GROUP)
        return self._("Title of group {0} is set to {1}.").format(*args)

    @catch_exceptions
    def set_chat_emoji(self, args: str) -> str:
        """ChatID emoji"""
        args = args.split(' ', 1)
        if len(args) < 2:
            return self._("User ID and emoji are required.")
        self.client.changeThreadEmoji(args[1], args[0])
        return self._("Emoji of group {0} is set to {1}.").format(*args)

    @catch_exceptions
    def set_member_nickname(self, args: str) -> str:
        """GroupID MemberID nickname"""
        args = args.split(' ', 2)
        if len(args) < 3:
            return self._("Group ID, member ID and are nickname required.")
        self.client.changeNickname(args[2], args[1], args[0])
        return self._("Nickname of {1} in {0} is set as {2}.").format(*args)



# coding=utf-8

"""
Authorize user and store session into the correct path.
"""

import getpass
import argparse
import pickle
import os
from gettext import translation
from pkg_resources import resource_filename

from fbchat import Client
from ehforwarderbot import coordinator, utils

from .__version__ import __version__
from . import FBMessengerChannel


def main():
    translation("efb_fb_messenger_slave",
                resource_filename("efb_fb_messenger_slave", "locale"),
                fallback=True).install()

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--profile", help=_("Target profile."), default="default")
    parser.add_argument("-i", "--instance", help=_("Target instance ID."), default=None)

    args = parser.parse_args()
    profile = args.profile
    coordinator.profile = profile
    channel_id = FBMessengerChannel.channel_id
    if args.instance:
        channel_id += f"#{args.instance}"
    path = utils.get_data_path(channel_id)

    print(_("EFB Facebook Messenger Slave Session Updater\n"
            "============================================"))
    print()
    print(_("You are running EFMS {0}.").format(__version__))
    print()
    print(_("This will update and overwrite your EFMS token by\n"
            "log into your account again manually."))
    print()
    print(_("You usually need to do this when you want to log into\n"
            "a new account, or when the previous session is expired."))
    print()
    print(_("This session is written to"
            "{0}").format(path))
    print()
    print(_("To continue, press Enter/Return."))

    input()

    email = input(_("Email: "))
    password = getpass.getpass(_("Password: "))
    client = Client(email, password)

    if not client.isLoggedIn():
        print(_("Log in failed. Please check the information above and try again."))
        exit(1)

    session_path = os.path.join(path, "session.pickle")

    with open(session_path, 'wb') as f:
        pickle.dump(client.getSession(), f)

    print(_("Your session has been successfully updated. It's stored at\n"
            "{0}").format(session_path))
    print(_("Your session cookies is as valuable as your account credential.\n"
            "Please keep them with equal care."))


if __name__ == '__main__':
    main()
